function [transformed, time] = GaussianMixture(data1,data2,vis)
## /**********************************************************************
##   Copyright 2013, Hanchen Xiong
##   email: hanchen.xiong@uibk.ac.at
##
##   This file is part of 3D Point Registration Benchmark Framework
##
##   This is free software: you can redistribute it and/or modify
##   it under the terms of the GNU Lesser General Public License as published by
##   the Free Software Foundation, either version 3 of the License, or
##   (at your option) any later version.
##
##   This code is distributed in the hope that it will be useful,
##   but WITHOUT ANY WARRANTY; without even the implied warranty of
##   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##   GNU Lesser General Public License for more details.
##
##   You can find a copy of the GNU Lesser General Public License
##   in <http://www.gnu.org/licenses/>.
##
## ***********************************************************************/

% asumme that we align data1 with data2
%--------------------------------------------------------------------------
% extract information from data1 and data2
%--------------------------------------------------------------------------
tic
num_1=size(data1,1);
num_2=size(data2,1);
original_data1=data1;
original_data2=data2;
%--------------------------------------------------------------------------
% normalized two point clouds in the same reference frameword
%--------------------------------------------------------------------------
data1_mean=mean(data1,1);   data2_mean=mean(data2,1);      % mean
data1=data1-repmat(data1_mean,num_1,1);                    % normalization
data2=data2-repmat(data2_mean,num_2,1); 
%--------------------------------------------------------------------------
% set up the kernel parameters
%--------------------------------------------------------------------------
sigma=2*mean(std([data1;data2],0,1));

%--------------------------------------------------------------------------
% gradient ascent iteration on SE(3) manifold
%--------------------------------------------------------------------------
cstep=0;
nstep=800;

current_R=eye(3);                              % intial rotation matrix
current_L=(mean(data2)-mean(data1))';          % intial translation vector
current_P=[current_R,current_L; 0 0 0 1];      % intial pose matrix

h_data1=[data1,ones(num_1,1)]';         % homegeous coordinate of subset points, size is 4 by D+1 
h_data2=[data2,ones(num_2,1)]';         % homegeous coordinate of data2 points, size3 is 4 by num_2;

while cstep<nstep
    cstep=cstep+1;
    scale=10/(num_1*num_2*(cstep^0.1));                % update scale at every iteration
    h_trans_data1=current_P*h_data1;     % transformed data1  points with current pose 
    gradient_w=zeros(1,3);               % gradient w.r.t w
    gradient_v=zeros(1,3);               % gradient w.r.t v
        
    cross_kernel=kernel_eval(h_trans_data1(1:3,:)',data2,3,sigma);
    cross_kernel=cross_kernel';
    FX=cross_kernel; 
    D1=h_trans_data1'.*(FX*ones(num_2,1)*ones(1,4))-FX*h_data2';
    
    for i=1:num_1
        gradient_w=gradient_w-D1(i,:)*[-Omega(h_trans_data1(1:3,i));zeros(1,3)]/sigma^2;
        gradient_v=gradient_v-D1(i,:)*[eye(3);zeros(1,3)]/sigma^2;
    end;
    if norm(gradient_w)<0.0001 && norm(gradient_v)<0.0001
        break;
    end;
    w=(gradient_w')*scale;
    v=(gradient_v')*scale;
    exp_Gamma=expm([Omega(w),v;0 0 0 0]);
    current_P=exp_Gamma*current_P;
end;
toc
time=toc;
current_transed=current_P*([data1,ones(num_1,1)]');
current_transed=current_transed(1:3,:)';
transformed=current_P*h_data1+repmat([data2_mean';1],1,num_1);
transformed=transformed(1:3,:)';

%--------------------------------------------------------------------------
% visualize the source and target point clouds for alignment
%-------------------------------------------------------------------------
if vis~=0
    figure;
    set(gcf,'position',[400,200,700,500]);
    subplot(1,2,1); 
    scatter3(original_data2(:,1),original_data2(:,2),original_data2(:,3),15,'filled','black'); hold on;
    scatter3(original_data1(:,1),original_data1(:,2),original_data1(:,3),20,'red');
    set(gca,'xticklabel',[]);
    set(gca,'yticklabel',[]);
    set(gca,'zticklabel',[]);
    axis equal; axis vis3d;
    grid off;
    view(60,40);

    subplot(1,2,2); 
    scatter3(data2(:,1),data2(:,2),data2(:,3),15,'filled','black'); hold on;
    scatter3(current_transed(:,1),current_transed(:,2),current_transed(:,3),20,'red');
    set(gca,'xticklabel',[]);
    set(gca,'yticklabel',[]);
    set(gca,'zticklabel',[]);
    axis equal; axis vis3d;
    grid off;
    view(60,40);
        % store the tranformed point cloud and corresponding estimated pose
end;

%     end;    
    %----------------------------------------------------------------------
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [m]=Omega(w)
if size(w)~=[3,1]
    disp('the input of function Omega is not in right form');
else
    m=zeros(3,3);
    m(1,2)=-w(3);
    m(1,3)=w(2);
    m(2,1)=w(3);
    m(2,3)=-w(1);
    m(3,1)=-w(2);
    m(3,2)=w(1);
end;
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [K]=kernel_eval(X1,X2,ktype,ipar1,ipar2)
% X1,X2: two data set in which each data is a row, usualy is training set
% and test set 
% ktype: the type of kernel(0-linear 1-polynomial 2-sigmoid 3 Gaussian) 
% ipar1,par2: parameters used in different kernels, e.g. ipar1 is sigma 
% in Gaussian 
%--------------------------------------------------------------------------
    mtrain=size(X1,1);
    mtest=size(X2,1);
  
    d1=sum(X1.^2,2);
    d2=sum(X2.^2,2);

    K=X2*X1'; 
    switch ktype
      case 0     % linear kernel
      case 1     % polynomial
        K=(K+ipar2).^ipar1;
      case 2     % sigmoid
        K=tanh(ipar1*K+ipar2);
      case 3     % Gaussian
        K=d2*ones(1,mtrain)+ones(mtest,1)*d1'-2*K;
        K=exp(-K/(2*(ipar1^2)));
    end
end

