function [transformed, time] = SoftAssign(data1,data2,vis)
## /**********************************************************************
##   Copyright 2013, Hanchen Xiong
##   email: hanchen.xiong@uibk.ac.at
##
##   This file is part of 3D Point Registration Benchmark Framework
##
##   This is free software: you can redistribute it and/or modify
##   it under the terms of the GNU Lesser General Public License as published by
##   the Free Software Foundation, either version 3 of the License, or
##   (at your option) any later version.
##
##   This code is distributed in the hope that it will be useful,
##   but WITHOUT ANY WARRANTY; without even the implied warranty of
##   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##   GNU Lesser General Public License for more details.
##
##   You can find a copy of the GNU Lesser General Public License
##   in <http://www.gnu.org/licenses/>.
##
## ***********************************************************************/

% asumme that we align data1 with data2
%--------------------------------------------------------------------------
% extract information from data1 and data2
%--------------------------------------------------------------------------
tic
num_1=size(data1,1);
num_2=size(data2,1);
original_data1=data1;
original_data2=data2;
%--------------------------------------------------------------------------
% normalized two point clouds in the same reference frameword
%--------------------------------------------------------------------------
data1_mean=mean(data1,1);   data2_mean=mean(data2,1);      % mean
data1=data1-repmat(data1_mean,num_1,1);                    % normalization
data2=data2-repmat(data2_mean,num_2,1); 

%--------------------------------------------------------------------------
% rescale two point clouds into a canonical size
%--------------------------------------------------------------------------
data1_std=std(data1,0,1);                              % standard deviation
data2_std=std(data2,0,1);

for d=1:3
    if data1_std(d)~=0
        data1(:,d)=data1(:,d)./data1_std(d);
    end;
    if data2_std(d)~=0
        data2(:,d)=data2(:,d)./data2_std(d);
    end;
end;

%--------------------------------------------------------------------------
% initial pose P
%--------------------------------------------------------------------------
current_R=eye(3);                              % intial rotation matrix
current_L=(mean(data2)-mean(data1))';          % intial translation vector
current_P=[current_R,current_L; 0 0 0 1];      % intial pose matrix

h_data1=[data1,ones(num_1,1)]';         % homegeous coordinate of data1  points, size is 4 by num_1 
h_data2=[data2,ones(num_2,1)]';         % homegeous coordinate of data2 points, size is 4 by num_2;
m=ones(num_1+1,num_2+1);                % intial correspondence matrix
h_data1=[h_data1, mean(h_data1,2)];
h_data2=[h_data2, mean(h_data2,2)];
%--------------------------------------------------------------------------
% initial other parameters
%--------------------------------------------------------------------------
alpha=0.03;
d_hat=0;
for p=1:num_1
    for q=1:num_2
        d_hat=d_hat+norm(h_data2(:,q)-h_data1(:,p)).^2; 
    end;
end;
d_hat=d_hat/(num_1*num_2);

beta=0.1/d_hat;
beta_f=1000/d_hat;
beta_r=1.4;
I_0=10;
I_1=10;
nstep=200;

%--------------------------------------------------------------------------
% gradient ascent iteration on SE(3) manifold
%--------------------------------------------------------------------------
while beta<beta_f
    beta=beta*beta_r;
    if beta>beta_f
        break;
    end;
    old_m=zeros(num_1+1,num_2+1);
    % ------------computing the corresponden matrix m----------------------
    for i=1:I_0
        if norm(old_m-m)<0.01
             break;
        end;
        for p=1:num_1+1
            for q=1:num_2+1
                distance=norm(h_data2(:,q)-current_P*h_data1(:,p)).^2;
                m(p,q)=exp(-beta*(distance-alpha));
            end;
        end;
        %-------------normalize the row and colum of the matrix m----------
        normalized_m=zeros(num_1+1,num_2+1);
        for k=1:I_1
            if norm(m-normalized_m)<0.01
                break;
            end;
            for p=1:num_1+1
                sum_row_p=sum(m(p,:));
                m(p,:)=m(p,:)/sum_row_p;
            end; 
            for q=1:num_2+1
                sum_col_q=sum(m(:,q));
                m(:,q)=m(:,q)/sum_col_q;
            end;
            normalized_m=m;
        end;

        old_m=m;
        %--------------comput the motion parameters-------------------------
        cstep=0;
        current_R=eye(3);                              % intial rotation matrix
        current_L=(mean(data2)-mean(data1))';          % intial translation vector
        current_P=[current_R,current_L; 0 0 0 1];      % intial pose matrix
        

        while cstep<nstep
          cstep=cstep+1;
          scale=1/(sqrt(d_hat)*num_1*num_2);     % update scale at every iteration
          h_trans_data1=current_P*h_data1;      % transformed subset points with current pose 
          gradient_w=zeros(1,3);                % gradient w.r.t w
          gradient_v=zeros(1,3);                % gradient w.r.t v
               
          FX=m;
          D1=h_trans_data1'.*(FX*ones(num_2+1,1)*ones(1,4))-FX*h_data2';
    
          for t=1:num_1+1
              gradient_w=gradient_w-D1(t,:)*[-Omega(h_trans_data1(1:3,t));zeros(1,3)];
              gradient_v=gradient_v-D1(t,:)*[eye(3);zeros(1,3)];
          end;
          if norm(gradient_w)<0.1 && norm(gradient_v)<0.1
              break;
          end;
          w=(gradient_w')*scale;
          v=(gradient_v')*scale;
          exp_Gamma=expm([Omega(w),v;0 0 0 0]);
          current_P=exp_Gamma*current_P;
      end;
   end;
end;
toc
time=toc;
current_transed=current_P*([data1(1:num_1,:),ones(num_1,1)]');
current_transed=current_transed(1:3,:)';
transformed=current_P*h_data1(:,1:num_1)+repmat([data2_mean';1],1,num_1);
transformed=transformed(1:3,:)';

%--------------------------------------------------------------------------
% visualize the source and target point clouds for alignment
%-------------------------------------------------------------------------
if vis~=0
    figure(1);
    set(gcf,'position',[400,200,700,500]);
    subplot(1,2,1); 
    scatter3(original_data2(:,1),original_data2(:,2),original_data2(:,3),20,'+','black'); hold on;
    scatter3(original_data1(:,1),original_data1(:,2),original_data1(:,3),20,'red');
    set(gca,'xticklabel',[]);
    set(gca,'yticklabel',[]);
    set(gca,'zticklabel',[]);
    axis equal; axis vis3d;
    grid off;
    view(60,40);
    
    subset_size=10; 
    permutation=randperm(num_1);    % random permutation of [1,num_1]
    idx=permutation(1:subset_size); % select the first subset_size indices (random selection)
    idx=sort(idx); 
    for i=1:subset_size
        [max_value,max_rho_i_colum]=max(m(idx(i),:));  % find the correspondence in the ith column
        point_data1=original_data1(idx(i),1:3);
        point_data2=original_data2(max_rho_i_colum,1:3); 
        point_combine=[point_data1;point_data2];
        plot3(point_combine(:,1),point_combine(:,2),point_combine(:,3),'green');
    end;
 
    subplot(1,2,2); 
    scatter3(data2(:,1),data2(:,2),data2(:,3),20,'+','black'); hold on;
    scatter3(current_transed(:,1),current_transed(:,2),current_transed(:,3),20,'red');
    set(gca,'xticklabel',[]);
    set(gca,'yticklabel',[]);
    set(gca,'zticklabel',[]);
    axis equal; axis vis3d;
    grid off;
    view(60,40);
        % store the tranformed point cloud and corresponding estimated pose
end;
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [m]=Omega(w)
if size(w)~=[3,1]
    disp('the input of function Omega is not in right form');
else
    m=zeros(3,3);
    m(1,2)=-w(3);
    m(1,3)=w(2);
    m(2,1)=w(3);
    m(2,3)=-w(1);
    m(3,1)=-w(2);
    m(3,2)=w(1);
end;
end

